/**
* HomeEasy Library
*
* Usage notes : 
*     By default the library is hooked up to a fixed set of pins (for the benefit of the interrupts).  On your Arduino, you should connect the transmitter data to pin 13 and the receiver data to pin 8 
*     There is a degree of configurability for the transmission pin.  We use PORTB, so you must choose a pin on that PORT (PB*) from http://www.arduino.cc/en/Hacking/PinMapping168
*     To configure the transmitter data pin, define HETXPIN to the pin of choice from PB*
*  2015-12-10: Ben Smeets
*              Modified receiver and sending sketch after sketch from https://github.com/bruce33/homeeasyhacking/tree/master/Arduino
*              Removed simpleprotocol handler 
*/
#include "HomeEasy.h"

// for backwards compatibility
#ifndef PINB0
  #define PINB0 PB0
#endif
#ifndef PINB5
  #define PINB5 PB5
#endif

#ifndef HETXPIN
	#define HETXPIN PINB5
#endif
#ifndef HERXPIN
	#define HERXPIN PINB0
#endif

#define HERXPIN PINB0
#define HETXPIN PINB5

#define TRANSMITTER_MESSAGE_COUNT 5

#define MESSAGE_TYPE_SIMPLE 0
#define MESSAGE_TYPE_ADVANCED 1

#define TIME_ON   250
#define TIME_ONE  TIME_ON
#define TIME_ZRO  (5*TIME_ON)
#define TIME_SYN  (10*TIME_ON)
#define TIME_PAUS (40*TIME_ON)

// variables used for receiving the messages
unsigned int pulseWidth = 0;
unsigned int latchStage = 0;
signed int bitCount = 0;
byte bit = 0;
byte prevBit = 0;

// variables for storing the data received
unsigned long receivedData = 0;

// variables for storing the data received
unsigned long sendData = 0;



// variables for sending messages
byte messageType;
unsigned int messageCount;
unsigned long nextlowtime;
unsigned int signal;

// result handlers
void (*HomeEasy::advancedProtocolHandler)(unsigned long, unsigned int, bool, bool) = NULL;



/**
 * Constructor
 */
HomeEasy::HomeEasy()
{
}

  	
/**
 * Initialise the system.
 * 
 * Enables the receiving of messages.
 */
void HomeEasy::init()
{
	// ensure the receiver pin is set for input
	DDRB &= ~_BV(HERXPIN);
	
	// disable PWM (default)
	TCCR1A = 0x00;
	
	// set prescaler to 1/8.  TCNT1 increments every 0.5 micro seconds
	// falling edge used as trigger
	TCCR1B = _BV(CS21);
	
	// enable input capture interrupt for timer 1
	TIMSK1 = _BV(ICIE1);
}


/**
 * Reconfigure the interrupts for sending a message.
 */
void HomeEasy::initSending()
{
	// reset counter
	TCNT1 = 0;
        // ensure the transmitter pin is set for output
	DDRB |= _BV(HETXPIN);
	
	// the value that the timer will count up to before firing the interrupt
	OCR1A = (pulseWidth * 2);

	// toggle OC1A on compare match
	TCCR1A = 0;//_BV(COM1A0);

	// CTC mode: top of OCR1A, immediate update of OCR1A, TOV1 flag set on MAX
	TCCR1B |= _BV(WGM12);

	// enable timer interrupt for timer 1, disable input capture interrupt
	TIMSK1 = _BV(OCIE1A);
}

/**
 * Register a handler for the advanced protocol messages.
 */
void HomeEasy::registerAdvancedProtocolHandler(void(*handler)(unsigned long, unsigned int, bool, bool))
{
	HomeEasy::advancedProtocolHandler = handler;
}


/**
 * The input interrupt handler.
 * 
 * This is where the message is received and decoded.
 */
ISR(TIMER1_CAPT_vect)
{
	// reset counter
	TCNT1 = 0;
	
	// get value of input compare register, divide by two to get microseconds
	pulseWidth = (ICR1 / 2);

	if(bit_is_clear(TCCR1B, ICES1))
	{	// falling edge was detected, HIGH pulse end
		if(latchStage == 1 && pulseWidth > 150 && pulseWidth < 400)
		{	// advanced protocol latch
			latchStage = 2;
		}
		else if(latchStage == 3 && (pulseWidth < 150 || pulseWidth > 500))
		{	// advanced protocol data out of timing range
			latchStage = 0;
			bitCount = 0;
			receivedData = 0;
		}
	}
	else
	{	// raising edge was detected, LOW pulse end
		
		if(latchStage == 0 && pulseWidth > (TIME_PAUS-520) && pulseWidth < (TIME_PAUS+1500))
		{	// pause between messages
			latchStage = 1;
		}
		else if(latchStage == 2 && pulseWidth > (TIME_SYN-700) && pulseWidth < (TIME_SYN+700))
		{	// advanced protocol latch
			latchStage = 3;
		}
		else if(latchStage == 3)
		{	// advanced protocol data
			
			if(pulseWidth > (TIME_ONE-100) && pulseWidth < (TIME_ONE+150))
			{
				bit = 1;
			}
			else if(pulseWidth > (TIME_ZRO-250) && pulseWidth < (TIME_ZRO+150))
			{
				bit = 0;
			}
			else
			{	// start over if the low pulse was out of range
				latchStage = 0;
				bitCount = 0;
                                receivedData = 0;
			}
			
			if(bitCount % 2 == 1)
			{
				if((prevBit ^ bit) == 0)
				{	// must be either 01 or 10, cannot be 00 or 11					
					latchStage = 0;
					bitCount = -1;
				}
				else
				{					
					receivedData <<= 1;
					receivedData |= prevBit;
				}
			}			
			prevBit = bit;
			bitCount++;
			
			if(bitCount == 64)
			{	// message is complete
//				if(HomeEasy::advancedProtocolHandler != NULL)
				{
                                  unsigned long sender = 0x3FFFFFF & (receivedData >> 6);
                                  bool command = 1^((receivedData >> 4) & 0x01);
                                  bool group   = 1^((receivedData >> 5) & 0x01);
                                  unsigned int recipient = (unsigned int)receivedData & 0xF;
                                  HomeEasy::advancedProtocolHandler(sender, recipient, command, group);
          
				}				
				latchStage = 0;
				bitCount = 0;
                                return;       
			}
		}
	}
	// toggle bit value to trigger on the other edge
	TCCR1B ^= _BV(ICES1);
}



/**
 * 
 */
void HomeEasy::sendAdvancedProtocolMessage(unsigned long s, unsigned int r, bool c, bool g)
{
	// disable all interrupts
	TIMSK1 = 0;
	
	// reset variables
	messageCount = 0;
	latchStage = 0;
	bitCount = 0;
	signal = 0; // we start with a low
	pulseWidth = 1000;
	
	// set data to transmit
	sendData = s;
	sendData = (sendData << 1) | g;
        sendData = (sendData << 1) | c;
        sendData = (sendData << 4) | r;
	
	// specify encoding
	messageType = MESSAGE_TYPE_ADVANCED;
	
	// start the timer interrupt
	initSending();
}


/**
 * The timer interrupt handler.
 * 
 * This is where the message is transmitted.
 * 
 * The timer interrupt is used to wait for the required length of time.  Each call of this
 * function toggles the output and determines the length of the time until the function is
 * called again.
 * 
 * Once the message has been transmitted this class will switch back to receiving.
 */
ISR(TIMER1_COMPA_vect)
{	
    if(messageType == MESSAGE_TYPE_ADVANCED)
    {
 		 if((signal==0) && (messageCount == TRANSMITTER_MESSAGE_COUNT))
		 {	// go back to receiving
			messageCount = 0;
			
			TCCR1A = 0x00;
			TCCR1B = 0x02;
			TIMSK1 = _BV(ICIE1);			
			return;
		}     

		if (signal==1) {
                  // end of signal=1
                  PORTB &= ~_BV(HETXPIN);
                   signal = 0;
                  pulseWidth = nextlowtime;
                }
                else { 
                  // end of signal=0, how long next low time
                  PORTB |= _BV(HETXPIN);
                  signal = 1;
                  if (latchStage == 0){
                    // sync
                    nextlowtime = TIME_SYN;
                    latchStage = 1;
                  }                 
		  else if(bitCount == 64)
		  {	// message finished			
			messageCount++;
			// pause
			nextlowtime = TIME_PAUS;
			latchStage = 0;
			bitCount = 0;
		  }                
		  else if (latchStage == 1)
		  {
			if(bitCount % 2 == 1)
			{	// every other bit is inverted
				bit = !bit;
			}
			else
			{	// data
				bit = ((sendData >> (31-(bitCount / 2))) & 1) ? 1: 0;
			}
                        nextlowtime = (bit) ? TIME_ONE : TIME_ZRO;
			bitCount++;
		  };
                  pulseWidth = TIME_ON;
                } // end if signal=0
	}
	
	// set the next delay
	OCR1A = (pulseWidth * 2);
}

